/*
 * Copyright (C) 2016 Väinö Helminen
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "args.h"
#include "mime.h"
#include "strutil.h"
#include "list.h"
#include "log.h"

const char *WEBROOT = NULL;
const char *SERVER_ADDR = NULL;
int SERVER_PORT = -1;
int THREAD_POOL_SIZE = -1;
const char *INDEX_LIST = NULL;

static inline void print_usage(FILE *file, const char *exe)
{
  fprintf(
    file,
    "Usage: %s [-h|--help] [-a x.x.x.x] [-p n] [-t n] [-i idx.htm] [-m mime.txt]"
#ifdef DEBUG
    " [-L A|E|W|I|D|T]"
#endif
    " [dir]\n"
    "Where:\n"
    "  -h, --help   Print usage\n"
    "  -a x.x.x.x   Set listen IPv4 address (default: %s)\n"
    "  -p n         Set listen TCP port (default: %d)\n"
    "  -t n         Set number of threads (default: %d)\n"
    "  -i idx.htm   Set default page(s) (default: %s)\n"
    "  -m mime.txt  Load MIME table from a file (default: built-in)\n"
#ifdef DEBUG
    "  -L A|E|W|I|D|T  Set logging level (default is %c)\n"
#endif
    "  dir          Set webroot (default: %s)\n",
    exe,
    DEFAULT_ADDR,
    DEFAULT_PORT,
    DEFAULT_THREAD_POOL_SIZE,
    DEFAULT_INDEX_FILES,
#ifdef DEBUG
    "AEWIDT"[DEFAULT_LOG_LEVEL],
#endif
    DEFAULT_WEBROOT);
}

/**
 * Build a nul (\0) separated list from a comma (, ; or :) separated
 * list.
 *
 * Whitespace is trimmed from both ends. The result ends with double
 * nul.
 *
 * You can print the items with:
 *   for (int i = 0; s[i];) {
 *     puts(&s[i]);
 *     while(s[i++]);
 *   }
 */
static const char* parse_index_list(const char *src)
{
  const size_t length = strlen(src);
  char *dst = malloc(length+2);
  if (!dst) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  memcpy(dst, src, length);
  dst[length] = dst[length+1] = '\0';

  bool skip = true;
  size_t i = 0, j = 0;
  for (; i < length; ++i) {
    int c = dst[i];
    if (skip && c <= ' ')
      continue;
    if (c == '\0' || c == ',' || c == ';' || c == ':') {
      for (;j > 0; --j) {
        c = dst[j-1];
        if (c == '\0' || c > ' ') break;
      }
      c = '\0';
      skip = true;
    } else {
      skip = false;
    }
    dst[j++] = c;
  }
  dst[j] = '\0';

  return dst;
}

void parse_args(int argc, char **argv)
{
  bool has_mime = false;
  int i;
  for (i = 1; i < argc; ++i) {
    if (strcmp("-h", argv[i]) == 0 ||
       strcmp("--help", argv[i]) == 0) {
      print_usage(stdout, argv[0]);
      exit(EXIT_SUCCESS);
    }

    if (strcmp("-a", argv[i]) == 0 && i+1 < argc && !SERVER_ADDR) {
      SERVER_ADDR = argv[++i];
      continue;
    }

    if (strcmp("-p", argv[i]) == 0 && i+1 < argc && SERVER_PORT < 0) {
      SERVER_PORT = atoi(argv[++i]);
      if (SERVER_PORT > 0 && SERVER_PORT <= 0xffff)
        continue;
    }

    if (strcmp("-t", argv[i]) == 0 && i+1 < argc && THREAD_POOL_SIZE < 0) {
      THREAD_POOL_SIZE = atoi(argv[++i]);
      if (THREAD_POOL_SIZE >= 0)
        continue;
    }

    if (strcmp("-i", argv[i]) == 0 && i+1 < argc && !INDEX_LIST) {
      trace("parse_index_list");
      INDEX_LIST = parse_index_list(argv[++i]);
      continue;
    }

    if (strcmp("-m", argv[i]) == 0 && i+1 < argc && !has_mime) {
      trace("load_mime_table");
      load_mime_table(argv[++i]);
      has_mime = true;
      continue;
    }

#ifdef DEBUG
    if (strcmp("-L", argv[i]) == 0 && i+1 < argc) {
      switch(argv[++i][0]) {
      case 'A':
        LOG_LEVEL = LOG_ASSERT;
        continue;
      case 'E':
        LOG_LEVEL = LOG_ERROR;
        continue;
      case 'W':
        LOG_LEVEL = LOG_WARN;
        continue;
      case 'I':
        LOG_LEVEL = LOG_INFO;
        continue;
      case 'D':
        LOG_LEVEL = LOG_DEBUG;
        continue;
      case 'T':
        LOG_LEVEL = LOG_TRACE;
        continue;
      default:
        break;
      }
    }
#endif

    if (!WEBROOT) {
      WEBROOT = argv[i];
      continue;
    }

    fprintf(stderr, "Invalid argument: %s\n", argv[i]);
    print_usage(stderr, argv[0]);
    exit(EXIT_FAILURE);
  }

  if (!SERVER_ADDR) {
    SERVER_ADDR = DEFAULT_ADDR;
  }

  if (SERVER_PORT < 0) {
    SERVER_PORT = DEFAULT_PORT;
  }

  if (THREAD_POOL_SIZE < 0) {
    THREAD_POOL_SIZE = DEFAULT_THREAD_POOL_SIZE;
  }

  if (!WEBROOT) {
    WEBROOT = DEFAULT_WEBROOT;
  }

  if (!INDEX_LIST) {
    trace("load_mime_table");
    INDEX_LIST = parse_index_list(DEFAULT_INDEX_FILES);
  }

  if (!has_mime) {
    trace("default_mime_table");
    default_mime_table();
  }
}
