/*
 * Copyright (C) 2016 Väinö Helminen
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hashtable.h"
#include "log.h"

typedef struct _Node {
  const void *key;
  const size_t hash;
  void* value;
  struct _Node* next;
} Node;

struct _Hashtable {
  size_t size;
  size_t used;
  Node **table;
  size_t (*hash)(const void*);
  int (*cmp)(const void*, const void*);
};

Hashtable* ht_alloc(const size_t size, size_t (*hash)(const void*), int (*cmp)(const void*, const void*))
{
  assert(hash);
  assert(cmp);

  Hashtable *ht = malloc(sizeof *ht);
  if(!ht) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }

  ht->size = size ? (size*4+2)/3 : 16;
  ht->used = 0;

  ht->table = calloc(ht->size, sizeof *ht->table);
  if(!ht->table) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }

  ht->hash = hash;
  ht->cmp = cmp;
  return ht;
}

static void ht_rehash(Hashtable *ht, const size_t size)
{
  assert(ht);
  assert(size > ht->used);

  Node **table = calloc(size, sizeof *ht->table);
  if(!table) {
      perror("malloc");
      exit(EXIT_FAILURE);
  }

  for(size_t i = 0; i < ht->size; ++i) {
    Node **list = &ht->table[i];
    Node *node = *list;
    *list = NULL;

    while(node) {
      Node *next = node->next;

      list = &table[node->hash % size];
      node->next = *list;
      *list = node;

      node = next;
    }
  }

  free(ht->table);
  ht->table = table;
  ht->size = size;
}

void ht_free(Hashtable **ht)
{
  assert(ht);
  assert(*ht);

  for(size_t i = 0; i < (*ht)->size; ++i) {
    Node **list = &(*ht)->table[i];
    Node *node = *list;
    *list = NULL;

    while(node) {
      Node *next = node->next;
      free(node);
      node = next;
    }
  }
  free((*ht)->table);
  (*ht)->table = NULL;

  free(*ht);
  *ht = NULL;
}

void* ht_set(Hashtable *ht, const void *key, void* value)
{
  assert(ht);

  const size_t hash = ht->hash(key);

  Node **list = &ht->table[hash % ht->size];
  Node *node = *list;
  while(node) {
    if(ht->cmp(key, node->key) == 0) {
      void *old = node->value;
      node->value = value;
      return old;
    }
    node = node->next;
  }

  node = malloc(sizeof *node);
  if(!node) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  node->key = key;
  *((size_t*)&node->hash) = hash;
  node->value = value;
  node->next = *list;
  *list = node;

  if(++ht->used > (ht->size*3)/4) {
    ht_rehash(ht, ht->size*2);
  }

  return NULL;
}

void* ht_get(const Hashtable *ht, const void *key)
{
  assert(ht);

  const size_t hash = ht->hash(key);
  Node *node = ht->table[hash % ht->size];
  while(node) {
    if(ht->cmp(key, node->key) == 0) {
      return node->value;
    }
    node = node->next;
  }
  return NULL;
}

void* ht_remove(Hashtable *ht, const void *key)
{
  assert(key);

  const size_t hash = ht->hash(key);
  Node **list = &ht->table[hash % ht->size];
  Node *node = *list;
  while(node && ht->cmp(key, node->key) != 0) {
    list = &node->next;
    node = node->next;
  }
  if(!node) {
    return NULL;
  }

  *list = node->next;
  node->next = NULL;
  void *value = node->value;
  free(node);

  if(--ht->used < ht->size/4) {
    ht_rehash(ht, ht->size/2);
  }

  return value;
}

size_t ht_hash_string(const void *p)
{
  if(!p) return -1;
  const char *s = p;
  size_t hash = 5381;
  int c;
  while((c = *s++)) {
    hash = ((hash << 5) + hash) ^ c;
  }
  return hash;
}

int ht_cmp_string(const void *a, const void *b)
{
  if(a == b) return 0;
  if(!a) return b ? 1 : 0;
  if(!b) return -1;
  return strcmp(a, b);
}
