/*
 * Copyright (C) 2016 Väinö Helminen
 */
#ifndef HASHTABLE_H_1IKA65NX
#define HASHTABLE_H_1IKA65NX

typedef struct _Hashtable Hashtable;

/**
 * Allocate a new hashtable.
 *
 * The hash function can be anything as long as it returns the same
 * value for the same key.
 *
 * The cmp function should return zero for equal keys and non-zero
 * otherwise. (@see strcmp().)
 *
 * @param size number of initial items
 * @param hash hash function for keys
 * @param cmp compare function for keys
 * @return a hashtable
 */
Hashtable *ht_alloc(const size_t size, size_t (*hash)(const void*), int (*cmp)(const void*, const void*));

/**
 * Free all structures required by the hashtable.
 *
 * Sets *ht to NULL, too.
 *
 * Note: ht_free() does NOT free keys or values!
 *
 * @param ht pointer to the hashtable
 */
void ht_free(Hashtable **ht);

/**
 * Set a value for a key.
 *
 * The hashtable is automatically resized if over 3/4 full.
 *
 * @param ht the hashtable
 * @param key the key
 * @param value the value
 * @return old value or NULL
 */
void* ht_set(Hashtable *ht, const void *key, void* value);

/**
 * Find a value corresponding to the key.
 *
 * @param ht the hashtable
 * @param key the key
 * @return value or NULL
 */
void* ht_get(const Hashtable *ht, const void *key);

/**
 * Remove key / value pair from the hashtable.
 *
 * The hashtable is automatically resized if less than 1/4 full.
 *
 * @param ht the hashtable
 * @param key the key
 * @return value or NULL
 */
void* ht_remove(Hashtable *ht, const void *key);

/**
 * A basic hash function for strings. Works with NULL keys, too.
 */
size_t ht_hash_string(const void*);

/**
 * A simple @see strcmp() based cmp function for strings. Works with
 * NULL keys, too.
 */
int ht_cmp_string(const void*, const void*);

#endif
