/*
 * Copyright (C) 2016 Väinö Helminen
 */
#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#ifdef __linux__
#  include <sys/sendfile.h>
#endif
#ifdef __WIN32__
#  include <windows.h>
#  include <winsock2.h>
#  include <ws2tcpip.h>
#  include <Mswsock.h>
#endif
#include "config.h"
#include "args.h"
#include "inet.h"
#include "request.h"
#include "strutil.h"
#include "strbuf.h"
#include "list.h"
#include "path.h"
#include "http.h"
#include "mime.h"
#include "thread.h"
#include "log.h"
#include "favicon.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif

/* Time format for logging and directory listings. */
#define TIMESTAMP_FORMAT "%Y-%d-%m %H:%M:%S"
#define TIMESTAMP_LENGTH 20

static SOCKET open_server(const char * const addr, const int port)
{
  assert(addr);
  assert(port > 0 && port <= 0xffff);

  struct sockaddr_in sockaddr;
  memset(&sockaddr, 0, sizeof sockaddr);
  sockaddr.sin_family = AF_INET;
#ifdef __WIN32__
  if((sockaddr.sin_addr.s_addr = inet_addr(addr)) == INADDR_NONE) {
    socketerror("inet_addr");
    exit(EXIT_FAILURE);
  }
#else
  if(inet_pton(AF_INET, addr, &sockaddr.sin_addr) < 0) {
    socketerror("inet_pton");
    exit(EXIT_FAILURE);
  }
#endif
  sockaddr.sin_port = htons(port);

  const SOCKET sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(sock == INVALID_SOCKET) {
    socketerror("socket");
    exit(EXIT_FAILURE);
  }

  if(bind(sock, (struct sockaddr*)&sockaddr, sizeof sockaddr) < 0) {
    socketerror("bind");
    exit(EXIT_FAILURE);
  }

  if(listen(sock, LISTEN_BACKLOG) < 0) {
    socketerror("listen");
    exit(EXIT_FAILURE);
  }

  debug("socket: %d", (int)sock);

  return sock;
}

static Request* wait_client(const SOCKET server_socket)
{
  assert(server_socket != INVALID_SOCKET);

  struct sockaddr_in client_addr;
  socklen_t client_addr_size = 0;

  trace("alloc_request");
  Request *request = alloc_request();
  assert(request);

  for(;;) {
    client_addr_size = sizeof client_addr;
    trace("accept");
    SOCKET sock = accept(server_socket, (struct sockaddr*)&client_addr, &client_addr_size);
    if(sock != INVALID_SOCKET) {
      trace("socket %d", (int)sock);
      request->socket = sock;
      break;
    }

    if(errno != EAGAIN) {
      socketerror("accept");
      exit(EXIT_FAILURE);
    }
  }

  char *addr = NULL;
#ifdef __WIN32__
  trace("inet_ntoa");
  addr = inet_ntoa(client_addr.sin_addr);
  trace("copy_string");
  addr = copy_string(addr);
#else
  addr = malloc(INET_ADDRSTRLEN);
  if(!addr) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  trace("inet_ntop");
  if(!inet_ntop(AF_INET, &client_addr.sin_addr, addr, INET_ADDRSTRLEN)) {
    socketerror("inet_ntop");
    exit(EXIT_FAILURE);
  }
#endif

  request->client_addr = addr;

  debug("address: %s:%d", addr, (int)client_addr.sin_port);

  return request;
}

static Request* read_request(Request *request)
{
  char buffer[REQUEST_BUFFER_SIZE];
  int org = 0;
  off_t offset = 0;
#ifdef __WIN32__
    fd_set readfds;
#endif
  while(offset < sizeof buffer - 1) {
#ifdef __WIN32__
    FD_ZERO(&readfds);
    FD_SET(request->socket, &readfds);
    trace("select");
    if(select(1, &readfds, NULL, NULL, NULL) < 0) {
      const int err = WSAGetLastError();
      switch(err) {
      case WSAEWOULDBLOCK:
        break;
      case WSAECONNRESET:
        goto exitrecv;
      default:
        win32error("select", err);
        goto exitrecv;
      }
    }
    if(!FD_ISSET(request->socket, &readfds))
      continue;
#endif

    trace("recv");
    ssize_t bytes = recv(request->socket, buffer + offset, sizeof buffer - 1 - offset, 0);
    if(bytes <= 0) {
      if(bytes < 0) {
#ifdef __WIN32__
        const int err = WSAGetLastError();
        switch(err) {
        case 0:
        case WSAEWOULDBLOCK:
          continue;
        case WSAECONNRESET:
          goto exitrecv;
        default:
          win32error("recv", err);
          goto exitrecv;
        }
#else
        socketerror("recv");
#endif
      }
      break;
    }
    org += bytes;

    // Strip \r and stop receiving when we find the empty line after
    // the headers. (Any data after that will be lost.)
    off_t i = -3, nl = 0;
    if(i < offset) i = -offset;
    for(off_t j = i; j < bytes; ++i, j++) {
      int c =  (unsigned char)buffer[offset+i];
      buffer[offset+j] = c;
      if(c == '\0') {
        offset = j;
        goto exitrecv;
      } else if(c == '\n') {
        if(++nl == 2) {
          offset += j + 1;
          goto exitrecv;
        }
      } else if((c < ' ' && c != '\t') || c == 127) {
        --j;
        --bytes;
      } else {
        nl = 0;
      }
    }
    offset += bytes;
  }
exitrecv:
  buffer[offset] = '\0';

  debug("request: %s", buffer);

  if(buffer[0]) {
    trace("copy_string");
    request->request = copy_string(buffer);
  }

  return request;
}

static char* local_timestamp(char *buf, const size_t bufsize, const time_t *timestamp)
{
  assert(buf);
  assert(timestamp);

  *buf = '\0';

  if(*timestamp != -1) {
#ifdef __WIN32__
    struct tm *tm = localtime(timestamp);
    if(!tm) {
      perror("localtime");
    } else {
      strftime(buf, bufsize, TIMESTAMP_FORMAT, tm);
    }
#else
    struct tm tm;
    if(!localtime_r(timestamp, &tm)) {
      perror("localtime_r");
    } else {
      strftime(buf, bufsize, TIMESTAMP_FORMAT, &tm);
    }
#endif
  }

  return buf;
}

static void finish_request(Request **request)
{
  assert(request);

  trace("close socket %d", (int)(*request)->socket);
  closesocket((*request)->socket);

  char timestamp[TIMESTAMP_LENGTH];
  trace("local_timestamp");
  local_timestamp(timestamp, sizeof timestamp, &(*request)->timestamp);

  char *buf = (*request)->request;
  if(buf) {
    for(int i = 0; buf[i]; ++i) {
      if(buf[i] == '\r' || buf[i] == '\n') {
        buf[i] = '\0';
        break;
      }
    }
  } else {
    buf = "";
  }

  fprintf(stdout, "%s [%s] \"%s\" %d %lu\n",
          (*request)->client_addr,
          timestamp,
          buf,
          (*request)->status,
          (unsigned long)(*request)->response_size);


  trace("free_request");
  free_request(request);
  assert(!*request);
}

static bool send_http_headers(Request *request)
{
  assert(request);

  StringBuffer *sb = strbuf_alloc(HEADERS_BUFFER_SIZE);
  assert(sb);

  strbuf_format(sb, "HTTP/1.0 %d %s\r\n",
                request->status, get_status_message(request->status));;

  strbuf_append(sb, "Connection: close\r\n");

  for(List *node = request->response_headers; node; node = node->next) {
    strbuf_append(sb, node->value);
    strbuf_append(sb, "\r\n");
  }
  {
    char timestamp[HTTP_DATE_LENGTH];
    strbuf_format(sb, "Date: %s\r\n", get_http_timestamp(timestamp, HTTP_DATE_LENGTH, request->timestamp));
  }
  strbuf_format(sb, "Server: %s/%s\r\n", PROGRAM, VERSION);
  strbuf_append(sb, "\r\n");

  debug("headers size: %lu", (unsigned long)sb->size);
  ssize_t obytes = send(request->socket, sb->buffer, sb->size, MSG_NOSIGNAL);
  debug("sent: %ld", (long)obytes);
  assert(obytes == sb->size);
  bool success = obytes == sb->size;

  trace("strbuf_free");
  strbuf_free(&sb);
  assert(!sb);

#ifdef LOG_INCLUDE_HEADERS_SIZE
  if(obytes > 0) {
    request->response_size += obytes;
  }
#endif

  trace("success: %s", success?"true":"false");
  return success;
}

static bool send_strbuf_response(Request *request, const StringBuffer *sb, const int status, const char *mime)
{
  assert(request);
  assert(sb);
  assert(status >= 100 && status <= 599);
  assert(mime);

  request->status = status;
  trace("set_content_length");
  set_content_length(request, sb->size);
  trace("set_content_type");
  set_content_type(request, mime);

  trace("send_http_headers");
  bool success = send_http_headers(request);

  if(success &&
     request->method && strcmp("HEAD", request->method) != 0) {
    debug("content size: %lu", (unsigned long)sb->size);
    ssize_t obytes = send(request->socket, sb->buffer, sb->size, MSG_NOSIGNAL);
    debug("sent: %ld", (long)obytes);
    if(obytes > 0) {
      request->response_size += obytes;
    }
    success = obytes == sb->size;
  }

  trace("success: %s", success?"true":"false");
  return success;
}

static bool send_synthetic(Request *request, const int status)
{
  assert(request);
  assert(status >= 100 && status <= 599);

  list_free(&request->response_headers, true);
  assert(!request->response_headers);

  const char *message = get_status_message(status);
  assert(message);

  char *safe = NULL;
  if(request->path) {
    safe = htmlescape(request->path);
  }

  StringBuffer *sb = strbuf_alloc(FILE_BUFFER_SIZE);
  assert(sb);

  strbuf_format(sb,
    "<!doctype html><html><head><title>%d %s</title><style>body{background:#fff;color:#000;}</style></head><body><h1><span style=\"background:#000;color:#fff;padding:10px;border-radius:50%;\">%d</span> %s</h1><p>%s</p></body></html>",
    status, message,
    status, message,
    safe ? safe : "");

  if(safe) {
    free(safe);
    safe = NULL;
  }

  trace("send_strbuf_response");
  bool success = send_strbuf_response(request, sb, status, "text/html;charset=UTF-8");

  strbuf_free(&sb);
  assert(!sb);

  trace("success: %s", success?"true":"false");
  return success;
}

static bool send_http_redirect(Request *request, const int status, const char *path)
{
  assert(request);
  assert(status >= 300 && status <= 399);
  assert(path);

  info("redirect: %s => %d %s", request->path, status, path);

  request->status = status;
  trace("add_response_header");
  add_response_header(request, "Location", path);

  trace("send_http_headers");
  return send_http_headers(request);
}

#if defined(__linux__)
static ssize_t sys_sendfile(SOCKET sock, const int fd, const size_t count)
{
  ssize_t obytes = sendfile(sock, fd, NULL, count);
  if(obytes < 0 && errno != EPIPE) {
    perror("sendfile");
    obytes = 0;
  }
  return obytes;
}

#elif defined(__FreeBSD__)
static ssize_t sys_sendfile(SOCKET sock, const int fd, const size_t count)
{
  off_t obytes = 0;
  trace("sendfile");
  int ret = sendfile(fd, sock, 0, count, NULL, &obytes, 0);
  if (ret < 0 && errno != EPIPE) {
    perror("sendfile");
  }
  return obytes;
}
#else
#  ifdef __WIN32__
static ssize_t sys_sendfile_fallback(SOCKET sock, const int fd, const size_t count)
#  else
static ssize_t sys_sendfile(SOCKET sock, const int fd, const size_t count)
#  endif
{
  char buf[FILE_BUFFER_SIZE];
  ssize_t total = 0, bytes, obytes;
  do {
    trace("read");
    bytes = read(fd, buf, sizeof buf);
    if(bytes <= 0) {
      if(bytes < 0) {
        perror("read");
      }
      break;
    }
    trace("read: %ld", (long)bytes);

    trace("send");
    obytes = send(sock, buf, bytes, MSG_NOSIGNAL);
    if(obytes == -1) {
      switch(errno) {
      default:
        perror("send");
      case EPIPE:
        break;
      }
      break;
    }
    trace("sent: %ld", (long)obytes);
    total += obytes;
  } while(bytes == obytes);
  return total;
}

#  ifdef __WIN32__
static bool sendfile_fallback = false;

static ssize_t sys_sendfile(SOCKET sock, const int fd, const size_t count)
{
  if(sendfile_fallback) {
    goto fallback;
  }

  trace("_get_osfhandle");
  const HANDLE hFile = (HANDLE)_get_osfhandle(fd);
  if(hFile == INVALID_HANDLE_VALUE) {
    perror("_get_osfhandle");
    return 0;
  }

  trace("TransmitFile");
  if(!TransmitFile(sock, hFile, count, 0, NULL, NULL, TF_USE_KERNEL_APC)) {
    const int err = WSAGetLastError();
    switch(err) {
    case 0:
    case WSAECONNRESET:
      break;
    case WSAEOPNOTSUPP:
      // TransmitFile() is not supported on Windows 9x
      sendfile_fallback = true;
      goto fallback;
    default:
      win32error("TransmitFile", err);
      return 0;
    }
  }

  return count;

fallback:
  return sys_sendfile_fallback(sock, fd, count);
}
#  endif

#endif

static bool send_file(Request *request, const DirEntry *entry)
{
  assert(request);
  assert(entry);

  debug("open: %s", entry->name);
  const int fd = open(entry->name, O_RDONLY | O_BINARY);
  if(fd == -1) {
    // TODO use errno to set status code (403/500/other?)
    perror("open");
    return send_synthetic(request, 403);
  }
  trace("fd: %d", fd);

  request->status = 200;
  trace("set_content_type");
  set_content_type(request, get_mime_type(entry->name));
  trace("set_content_length");
  set_content_length(request, entry->size);
  trace("set_last_modified");
  set_last_modified(request, entry->lastmod);

  bool success = send_http_headers(request);
  if(success &&
     strcmp("HEAD", request->method) != 0) {
    trace("sys_sendfile");
    ssize_t obytes = sys_sendfile(request->socket, fd, entry->size);
    debug("sent: %ld", (long)obytes);
    if(obytes > 0) {
      request->response_size += obytes;
    }
    success = entry->size == obytes;
  }

  trace("close: %d", fd);
  if(close(fd)) {
    perror("close");
  }

  trace("success: %s", success?"true":"false");
  return success;
}


static void send_directory_listing(Request *request, const DirEntry *entry)
{
  if(entry->lastmod && entry->lastmod != -1) {
    char timestamp[HTTP_DATE_LENGTH];
    add_response_header(request, "Last-Modified", get_http_timestamp(timestamp, HTTP_DATE_LENGTH, entry->lastmod));
  }

  StringBuffer *sb = strbuf_alloc(FILE_BUFFER_SIZE);
  assert(sb);

  {
    trace("htmlescape");
    char *safe = htmlescape(request->path);
    assert(safe);
    trace("strbuf_format");
    strbuf_format(sb,
      "<!doctype html><html><head><title>%s</title><style>body{background:#fff;color:#000;}th{text-align:left;border-bottom:1px solid #999;}tfoot td{border-top:1px solid;color:#999;}.s{text-align:right;}</style></head><body><table><tr><th colspan=\"3\">%s</th></thead><tbody>",
      safe, safe);
    free(safe);
    safe = NULL;
  }

  if(entry->name_length > 1) {
    trace("strbuf_append");
    strbuf_append(sb, "<tr><td><a href=\"../\">../</a></td></tr>");
  }

  DirEntry **entries = NULL;
  trace("list_directory");
  size_t size = list_directory(entry->name, &entries);
  debug("entries: %lu", (unsigned long)size);
  if(size) {
    assert(entries);
    assert(entries[0]);
    for(int i = 0; i < size; ++i) {
      char *safe, *url;
#ifdef __WIN32__
      {
        trace("convert_windows_1252_to_utf8");
	char *tmp = convert_windows_1252_to_utf8(entries[i]->name);
        trace("htmlescape");
	safe = htmlescape(tmp);
        trace("urlencode");
        url = urlencode(tmp);
	free(tmp);
        tmp = NULL;
      }
#else
      trace("htmlescape");
      safe = htmlescape(entries[i]->name);
      trace("urlencode");
      url = urlencode(entries[i]->name);
#endif
      assert(safe);
      assert(url);

      if(entries[i]->type == DIRENTRY_TYPE_DIR) {
        trace("strbuf_format");
        strbuf_format(sb,
          "<tr><td><a href=\"%s/\">%s/</a></td></tr>",
          url, safe);
      } else {
        char timestamp[TIMESTAMP_LENGTH];
        trace("local_timestamp");
        local_timestamp(timestamp, sizeof timestamp, &entries[i]->lastmod);

        trace("strbuf_format");
        strbuf_format(sb,
          "<tr><td><a href=\"%s\" rel=\"nofollow\">%s</a></td><td class=\"s\">%lu</td><td>%s</td></tr>",
          url, safe,
          (unsigned long)entries[i]->size,
          timestamp);
      }
      free(url);
      url = NULL;
      free(safe);
      safe = NULL;
    }

    trace("free_direntries");
    free_direntries(&entries);
  }
  assert(!entries);

  trace("strbuf_format");
  strbuf_format(sb,
    "</tbody><tfoot><tr><td colspan=\"3\">%s v%s</td></tr></foot></table></body></html>",
    PROGRAM, VERSION);

  trace("send_strbuf_response");
  send_strbuf_response(request, sb, 200, "text/html;charset=UTF-8");

  trace("strbuf_free");
  strbuf_free(&sb);
  assert(!sb);
}


static void send_directory(Request *request, const DirEntry *entry)
{
  assert(request);
  assert(entry);

  if(entry->name[0] != '.' || entry->name[1] != '\0' ||
     request->path[0] != '/' || request->path[1] != '\0') {
    char tmp[entry->name_length + 1 + 2];
    trace("sprintf");
    int ssize = sprintf(tmp, "/%s/", entry->name);
    if(ssize < 0) {
      perror("sprintf");
      exit(EXIT_FAILURE);
    }
    assert(ssize > 0 && ssize < sizeof tmp);

    if(strcmp(request->path, tmp) != 0) {
      trace("send_http_redirect");
      send_http_redirect(request, 301, tmp);
      return;
    }
  }

  for (size_t i = 0; INDEX_LIST[i]; ) {
    const char *index = INDEX_LIST + i;
    const size_t index_length = strlen(index);
    i += index_length + 1;

    char filename[entry->name_length + index_length + 2];

    trace("sprintf");
    int ssize = sprintf(filename, "%s/%s", entry->name, index);
    if (ssize < 0) {
      perror("sprintf");
      exit(EXIT_FAILURE);
    }
    assert(ssize > 0 && ssize < sizeof filename);

    struct stat sb;
    trace("stat");
    if(!stat(filename, &sb) && S_ISREG(sb.st_mode)) {
      trace("alloc_direntry");
      DirEntry *fileEntry = alloc_direntry(filename);
      assert(fileEntry);
      fileEntry->type = DIRENTRY_TYPE_FILE;
      fileEntry->size = sb.st_size;
      fileEntry->lastmod = sb.st_mtime;

      trace("send_file");
      send_file(request, fileEntry);
      trace("free");
      free(fileEntry);
      return;
    }
  }

  trace("send_directory_listing");
  send_directory_listing(request, entry);
}

static bool send_favicon(Request *request)
{
  assert(request);

  request->status = 200;
  trace("set_content_length");
  set_content_length(request, FAVICON_SIZE);
  trace("set_content_type");
  set_content_type(request, get_mime_type(FAVICON_PATH));

  trace("send_http_headers");
  bool success = send_http_headers(request);
  if(success) {
    trace("send");
    ssize_t obytes = send(request->socket, FAVICON, FAVICON_SIZE, MSG_NOSIGNAL);
    if(obytes >= 0) {
      request->response_size += obytes;
      success = obytes == FAVICON_SIZE;
    } else {
      perror("send");
    }
  }
  trace("success: %s", success?"true":"false");
  return success;
}


void request_handler(void *sockp)
{
  const SOCKET server_socket = *((SOCKET*)sockp);
  for(;;) {
    Request *request = wait_client(server_socket);
    assert(request);
    if(!request) {
      error("die");
      exit(EXIT_FAILURE);
    }

    if(parse_request(read_request(request))) {
      trace("send_synthetic");
      send_synthetic(request, 400);

    } else if(strcmp("GET", request->method) == 0 ||
              strcmp("HEAD", request->method) == 0) {
      trace("GET or HEAD method");
#ifdef __WIN32__
      {
        char *tmp = convert_utf8_to_windows_1252(request->path);
        free(request->path);
        request->path = tmp;
        debug("windows-1252: %s", tmp);
      }
#endif
      DirEntry *entry = resolve_path(request->path);
      if(entry) {
        switch(entry->type) {
        case DIRENTRY_TYPE_DIR:
          trace("send_directory");
          send_directory(request, entry);
          break;
        case DIRENTRY_TYPE_FILE:
          trace("send_file");
          send_file(request, entry);
          break;
        default:
          trace("send_synthetic");
          send_synthetic(request, 422);
          break;
        }
        free(entry);
      } else if(strcmp(FAVICON_PATH, request->path) == 0) {
        trace("send_favicon");
        send_favicon(request);
      } else {
        trace("send_synthetic");
        send_synthetic(request, 404);
      }
    } else {
      trace("send_synthetic");
      send_synthetic(request, 405);
    }

    finish_request(&request);
    assert(!request);
  }
}

#ifdef __WIN32__
static char* realpath(const char *path, char* ignored)
{
  char buf[MAX_PATH];
  trace("GetFullPathName");
  if(!GetFullPathName(".", sizeof buf, buf, NULL))
    return NULL;
  trace("copy_string");
  return copy_string(buf);
}
#endif

int main(int argc, char **argv)
{
  tzset();

  fprintf(stdout,
          "%s v%s (%s %s) %s\n",
          PROGRAM, VERSION, __DATE__, __TIME__, COPYRIGHT);

  trace("parse_args");
  parse_args(argc, argv);

  trace("chdir");
  if(chdir(WEBROOT)) {
    perror("chdir");
    return EXIT_FAILURE;
  }
  trace("realpath");
  WEBROOT = realpath(CWD, NULL);
  if(!WEBROOT) {
    perror("realpath");
    return EXIT_FAILURE;
  }

#ifdef __WIN32__
  trace("load_favicon_ico");
  load_favicon_ico();

  trace("winsock_init");
  winsock_init();
#endif

  trace("open_server");
  SOCKET server_socket = open_server(SERVER_ADDR, SERVER_PORT);
  assert(server_socket != INVALID_SOCKET);

  fprintf(stdout,
          "Listening at %s:%d\n"
          "Serving files from %s\n",
          SERVER_ADDR, SERVER_PORT,
          WEBROOT);

  for(int i = 1; i < THREAD_POOL_SIZE; ++i) {
    trace("thread_create");
    if(thread_create(request_handler, &server_socket)) {
      break;
    }
  }

  trace("request_handler");
  request_handler(&server_socket);

  return EXIT_FAILURE;
}
