/*
 * Copyright (C) 2016 Väinö Helminen
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "mime.h"
#include "args.h"
#include "strutil.h"
#include "hashtable.h"
#include "log.h"

static const char * const MIME_TYPES[] = {
  "BMP",	"image/bmp",
  "CSS",	"text/css;charset=UTF-8",
  "GIF",	"image/gif",
  "HTM",	"text/html;charset=UTF-8",
  "HTML",	"text/html;charset=UTF-8",
  "ICO",	"image/x-icon",
  "JPE",	"image/jpeg",
  "JPEG",	"image/jpeg",
  "JPG",	"image/jpeg",
  "JS",	"application/javascript;charset=UTF-8",
  "JSON",	"application/json",
  "LICENSE",	"text/plain;charset=UTF-8",
  "LOG",	"text/plain",
  "PDF",	"application/pdf",
  "PNG",	"image/png",
  "README",	"text/plain;charset=UTF-8",
  "SVG",	"image/svg+xml",
  "TEXT",	"text/plain;charset=UTF-8",
  "TXT",	"text/plain;charset=UTF-8",
  "WEBM",	"video/webm",
  "WEBP",	"image/webp",
  "XHTML",	"application/xhtml+xml",
  "XML",	"text/xml",
  NULL,	"application/octet-stream", /* default MIME type for unknown files */
};

static Hashtable* mime_table = NULL;

void default_mime_table()
{
  const size_t size = sizeof MIME_TYPES / sizeof *MIME_TYPES;
  mime_table = ht_alloc(size/2, ht_hash_string, ht_cmp_string);
  for(size_t i = 0; i < size; i += 2) {
    ht_set(mime_table, MIME_TYPES[i], (char*)MIME_TYPES[i+1]);
  }
}

/**
 * Line format is (as a pseudo-regex):
 *   ^[.]*(ext)\s*\t\s*(mime/type)s*(\t.*)?$
 *
 * If (ext) is empty (e.g. the line starts with whitespace), that line
 * defines the default MIME type. (mime/type) must not be
 * empty. Redefining extensions is allowed but recommended.
 *
 * (ext) should start with a dot (.) if you do not intent it to match
 * only full filenames. Matching is case insensitive.
 *
 * (mime/type) can contain a charset definition after the type as long
 * as there are no tabs in between.
 *
 * You can add comments or other data that will be ignored after a
 * seconds tab.
 *
 * Note: Content of buf will be modified.
 */
static ssize_t parse_mime_table(Hashtable *mime, Hashtable *dedup, char *buf, size_t size)
{
  off_t off = 0;
  for (; off < size; ) {
    off_t eol = off;
    for(; eol < size && buf[eol] != '\n' && buf[eol] != '\r'; ++eol);
    if(eol >= size)
      break;

    {
      buf[eol] = '\0';
      char *tmp = trim(buf+off);
debug("'%s'", tmp);
      if (*tmp == '\0' || *tmp== '#')
        goto skip;
    }

    off_t sep1 = off;
    for(; sep1 < eol && buf[sep1] != '\t' && buf[sep1] != '#'; ++sep1);
    buf[sep1] = '\0';
    trace("trim");
    char *ext = trim(buf+off);

    off_t sep2 = ++sep1;
    for(; sep2 < eol && buf[sep2] != '\t' && buf[sep2] != '#'; ++sep2);
    buf[sep2] = '\0';

    trace("trim");
    char *type = trim(buf + sep1);
    if (type[0] == '\0')
      goto invalid;

    if(*ext != '\0') {
      trace("copy_toupper");
      ext = copy_toupper(ext);
      trace("ht_get");
      char *tmp = ht_get(dedup, ext);
      if(tmp) {
        trace("free");
        free(ext);
        ext = tmp;
      } else {
        trace("ht_set");
        ht_set(dedup, ext, ext);
      }
    } else {
      ext = NULL;
    }

    {
      trace("ht_get");
      char *tmp = ht_get(dedup, type);
      if(tmp) {
        type = tmp;
      } else {
        trace("copy_string");
        type = copy_string(type);
        trace("ht_set");
        ht_set(dedup, type, type);
      }
    }

    ht_set(mime, ext, type);

    debug("%s -> %s", ext, type);

  skip:
    off = eol + 1;
  }

  return off;

invalid:
  error("invalid MIME table");
  exit(EXIT_FAILURE);
}

void load_mime_table(const char *filename)
{
  trace("open");
  const int fd = open(filename, O_RDONLY);
  if (fd == -1) {
    perror("open");
    exit(EXIT_FAILURE);
  }

  trace("ht_alloc");
  mime_table = ht_alloc(128, ht_hash_string, ht_cmp_string);

  trace("ht_alloc");
  Hashtable *dedup = ht_alloc(128, ht_hash_string, ht_cmp_string);

  size_t size = 0;
  char buf[100];
  for (;;) {
    trace("read");
    ssize_t bytes = read(fd, buf + size, sizeof buf - 1 - size);
    if (bytes <= 0) {
      if (bytes < 0) {
        perror("read");
      }
      break;
    }
    trace("read: %ld\n", (long)bytes);

    size += bytes;
    buf[size] = '\0';

    trace("parse_mime_table");
    bytes = parse_mime_table(mime_table, dedup, buf, size);
    if (bytes < size) {
      if (bytes <= 0) break;
      size -= bytes;
      memmove(buf, buf + bytes, size);
    } else {
      size = 0;
    }
  }

  if (size) {
    buf[size] = '\n';
    parse_mime_table(mime_table, dedup, buf, size+1);
  }

  trace("close");
  if (close(fd)) {
    perror("close");
  }

  trace("ht_free");
  ht_free(&dedup);
  assert(!dedup);
}


const char* get_mime_type(const char* filename)
{
  {
    const char *tmp = rstrstr(filename, "/");
    if (tmp) {
      filename = tmp + 1;
    }
  }

  for (;;) {
    trace("copy_toupper");
    char *upper = copy_toupper(filename);

    trace("ht_get");
    const char *type = ht_get(mime_table, upper);

    trace("free");
    free(upper);
    upper = NULL;

    if (type)
      return type;

    trace("strstr");
    const char *tmp = strstr(filename, ".");
    if (!tmp)
      return "application/octect-stream";

    filename = tmp + 1;
  }
}
