/*
 * Copyright (C) 2016 Väinö Helminen
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#include "path.h"
#include "strutil.h"
#include "list.h"

DirEntry* alloc_direntry(const char *name)
{
  size_t len = strlen(name);
  DirEntry *entry = malloc(sizeof *entry + len + 1);
  if(!entry) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  entry->type = DIRENTRY_TYPE_UNDEFINED;
  entry->size = 0;
  entry->lastmod = 0;
  entry->name_length = len;
  memcpy(entry->name, name, len + 1);
  return entry;
}

static List* list_dir_list(const char *path)
{
  errno = 0;
  DIR *dir = opendir(path);
  if(!dir) {
    switch(errno) {
    case EACCES:
      break;
    case ENOENT:
      break;
    default:
      perror("opendir");
    }
    return NULL;
  }

  List *list = NULL;
  while(1) {
    errno = 0;
    struct dirent *ent = readdir(dir);
    if(!ent) {
      if(errno) {
        perror("readdir");
      }
      break;
    }

    if(ent->d_name[0] == '.' ||
       ent->d_name[strlen(ent->d_name)-1] == '~')
      continue;

    short type = DIRENTRY_TYPE_UNDEFINED;
    switch(ent->d_type) {
    case DT_DIR:
      type = DIRENTRY_TYPE_DIR;
      break;
    case DT_REG:
      type = DIRENTRY_TYPE_FILE;
      break;
    case DT_UNKNOWN:
      break;
    default:
      continue;
    }

    size_t size = 0;
    time_t mtime = 0;

    if(type != DIRENTRY_TYPE_DIR) {
      char *tmp = NULL;
      alloc_sprintf(&tmp, "%s/%s", path, ent->d_name);

      errno = 0;
      struct stat sb;
      if(stat(tmp, &sb)) {
        perror("stat");
        free(tmp);
        break;
      }
      free(tmp);

      if(S_ISDIR(sb.st_mode)) {
        type = DIRENTRY_TYPE_DIR;
      } else if(S_ISREG(sb.st_mode)) {
        type = DIRENTRY_TYPE_FILE;
        size = sb.st_size;
        mtime = sb.st_mtime;
      } else {
        continue;
      }
    }

    DirEntry *entry = alloc_direntry(ent->d_name);
    entry->type = type;
    entry->size = size;
    entry->lastmod = mtime;
    list_push(&list, entry);
  }

  closedir(dir);

  return list;
}

static size_t list_dir_array(const char *path, DirEntry ***array)
{
  *array = NULL;

  List *list = list_dir_list(path);
  if(!list) {
    return 0;
  }

  const size_t size = list_length(list);

  *array = malloc(sizeof ***array * (size+1));
  if(!*array) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  List *node = list;
  for(int i = 0; i < size; ++i) {
    (*array)[i] = node->value;
    node = node->next;
  }
  (*array)[size] = NULL;

  list_free(&list, false);

  return size;
}

static int direntry_cmp(const void *p1, const void *p2) {
  const DirEntry *e1 = *(DirEntry**)p1, *e2 = *(DirEntry**)p2;
  if(e1->type == DIRENTRY_TYPE_DIR) {
    if(e2->type != DIRENTRY_TYPE_DIR) {
      return -1;
    }
  } else if(e2->type == DIRENTRY_TYPE_DIR) {
    return 1;
  }
  return strcmp(e1->name, e2->name);
}

static inline void sort_dir_list_array(DirEntry **entries, const size_t size)
{
  qsort(entries, size, sizeof *entries, direntry_cmp);
}

size_t list_directory(const char *path, DirEntry ***entries)
{
  const size_t size = list_dir_array(path, entries);
  if(size > 1) {
    sort_dir_list_array(*entries, size);
  }
  return size;
}

void free_direntries(DirEntry ***entries)
{
  if(!*entries) return;

  for(DirEntry **entry = *entries; *entry; ++entry) {
    free(*entry);
    *entry = NULL;
  }

  free(*entries);
  *entries = NULL;
}

static List* split_path(const char *path)
{
  if(!path) return NULL;

  List *list = NULL;

  int i, j;
  for(i = 0, j = 0; ; ++j) {
    if(path[j] && path[j] != '/' && path[j] != '\\')
      continue;

    if(i != j && (i+1 != j || path[i] != '.')) {
      if(i+2 == j && path[i] == '.' && path[i+1] == '.') {
        free(list_pop(&list));
      } else {
        list_push(&list, copy_substring(path, i, j));
      }
    }

    if(!path[j]) break;

    i = j+1;
  }

  list_reverse(&list);

  return list;
}

DirEntry* resolve_path(const char *inpath)
{
  struct stat sb;

  List* list = split_path(inpath);

  if(!list) {
    errno = 0;
    if(stat(CWD, &sb)) {
      if(errno != ENOENT) {
        perror("stat");
      }
      return NULL;
    }

    DirEntry *entry = alloc_direntry(CWD);
    entry->type = DIRENTRY_TYPE_DIR;
    entry->size = sb.st_size;
    entry->lastmod = sb.st_mtime;
    return entry;
  }

  char *path = NULL;
  {
    size_t length = sizeof CWD;
    for(List *node = list; node ; node = node->next) {
      length += strlen((char*)node->value) + 1;
    }
    path = malloc(length);
    if(!path) {
      perror("malloc");
      exit(EXIT_FAILURE);
    }
  }

  size_t pos = sprintf(path, CWD);

  for(List *node = list; node; node = node->next) {
    path[pos++] = '/';

    const char *name = node->value;
    size_t len = strlen(name);
    memcpy(path + pos, name, len+1);
    pos += len;

    errno = 0;
    if(stat(path, &sb)) {
      if(errno != ENOENT) {
        perror("stat");
      }
      free(path);
      path = NULL;
      break;
    }

    if(!S_ISDIR(sb.st_mode)) {
#ifdef __WIN32__
      /* Catch CON, PRN, AUX, NUL; COM?, LPT?, etc. */
      if(sb.st_mtime == -1) {
	free(path);
	path = NULL;
      }
#endif
      break;
    }
  }

  list_free(&list, true);

  if(!path) {
    return NULL;
  }

  DirEntry *entry = alloc_direntry(path+2);
  if(S_ISDIR(sb.st_mode)) {
    entry->type = DIRENTRY_TYPE_DIR;
  } else if(S_ISREG(sb.st_mode)) {
    entry->type = DIRENTRY_TYPE_FILE;
  }
  entry->size = sb.st_size;
  entry->lastmod = sb.st_mtime;

  free(path);
  path = NULL;

  return entry;
}
