/*
 * Copyright (C) 2016 Väinö Helminen
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "request.h"
#include "strutil.h"
#include "http.h"
#include "log.h"

Request* alloc_request(void)
{
  Request *request = NULL;
  request = malloc(sizeof *request);
  if(!request) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }

  request->timestamp = time(NULL);
  if(request->timestamp == -1) {
    perror("time");
  }

  request->socket = -1;

  request->client_addr = NULL;
  request->request = NULL;
  request->method = NULL;
  request->path = NULL;
  #if 0
  request->query = NULL;
  #endif

  request->status = 599; // It's a bug if no-one changes this
  request->response_size = 0;
  request->response_headers = NULL;

  return request;
}

void free_request(Request **request)
{
  assert(request);
  assert(*request);

  (*request)->socket = -1;
  if((*request)->client_addr) {
    free((*request)->client_addr);
    (*request)->client_addr = NULL;
  }
  if((*request)->request) {
    free((*request)->request);
    (*request)->request = NULL;
  }
  if((*request)->method) {
    free((*request)->method);
    (*request)->method = NULL;
  }
  if((*request)->path) {
    free((*request)->path);
    (*request)->path = NULL;
  }
  #if 0
  if((*request)->query) {
    free((*request)->query);
    (*request)->query = NULL;
  }
  #endif

  (*request)->status = 599;
  (*request)->response_size = 0;

  list_free(&(*request)->response_headers, true);

  free(*request);
  *request = NULL;
}

int parse_request(Request *request)
{
  const char *req = request->request;
  if(!req) {
    return -1;
  }

  int i = 0, j;

  for(j = 0; j < 8; ++j) {
    if(!req[j] || isspace(req[j])) {
      request->method = copy_substring(req, i, j);
      break;
    }
  }
  if(!request->method) {
    info("no method");
    return 1;
  }

  debug("method: %s", request->method);

  for(i = j; isspace(req[i]); ++i);

  char *path = NULL;
  for(j = i; ; ++j) {
    if(!req[j] || isspace(req[j]) || req[j] == '?') {
      path = copy_substring(req, i, j);
      break;
    }
  }
  if(!path) {
    debug("no path");
    return 2;
  }
  if(path[0] != '/') {
    debug("invalid path: %s", path);
    free(path);
    return 3;
  }

  trace("urldecode");
  request->path = urldecode(path);
  free(path);

  debug("path: %s", request->path);

  #if 0
  if(req[j] == '?') {
    for(i = j; ; ++j) {
      if(!req[j] || isspace(req[j]) || req[j] == '?') {
        request->query = copy_substring(req, i, j);
        break;
      }
    }
  }
  #endif

  return 0;
}

void set_content_type(Request *request, const char *mime)
{
  assert(request);
  assert(mime);
  trace("add_response_header");
  add_response_header(request, "Content-Type", mime);
}

void set_content_length(Request *request, size_t size)
{
  assert(request);
  trace("add_response_header_format");
  add_response_header_format(request, "Content-Length", "%lu", (unsigned long)size);
}

void set_last_modified(Request *request, time_t lastmod)
{
  assert(request);
  if(lastmod != 0 && lastmod != -1) {
    char timestamp[HTTP_DATE_LENGTH];
    trace("get_http_timestamp");
    get_http_timestamp(timestamp, HTTP_DATE_LENGTH, lastmod);
    trace("add_response_header");
    add_response_header(request, "Last-Modified", timestamp);
  }
}

void add_response_header(Request *request, const char *header, const char *value)
{
  assert(request);
  assert(header);
  assert(header[0]);
  assert(value);
  assert(value[0]);

  char *line = malloc(strlen(header) + strlen(value) + 5);
  if(!line) {
    perror("malloc");
    return;
  }

  if(sprintf(line, "%s: %s", header, value) < 0) {
    perror("sprintf");
    free(line);
    return;
  }

  debug(line);

  list_push(&request->response_headers, line);
}

void add_response_header_format(Request *request, const char *header, const char *format, ...)
{
  assert(request);
  assert(header);
  assert(format);

  va_list ap;
  va_start(ap, format);
  char *value = NULL;
  trace("alloc_vsprintf");
  alloc_vsprintf(&value, format, ap);
  va_end(ap);

  trace("add_response_header");
  add_response_header(request, header, value);

  free(value);
  value = NULL;
}
