/*
 * Copyright (C) 2016 Väinö Helminen
 */
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "strutil.h"
#include "log.h"

size_t alloc_vsprintf(char **buf, const char *format, va_list ap)
{
  *buf = NULL;

  va_list ap2;
  va_copy(ap2, ap);
  int size = vsnprintf(NULL, 0, format, ap2);
  if(size < 0) {
    perror("vsnprintf");
    exit(EXIT_FAILURE);
  }
  va_end(ap2);

  *buf = malloc(++size);
  if(!*buf) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }

  size = vsnprintf(*buf, size, format, ap);
  if(size < 0) {
    perror("vsnprintf");
    exit(EXIT_FAILURE);
  }

  return size;
}

size_t alloc_sprintf(char **buf, const char *format, ...)
{
  va_list ap;
  va_start(ap, format);
  size_t len = alloc_vsprintf(buf, format, ap);
  va_end(ap);
  return len;
}

char* copy_string(const char *src)
{
  size_t len = strlen(src) + 1;
  char *dst = malloc(len);
  if(!dst) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  return memcpy(dst, src, len);
}

char* copy_substring(const char *src, int from, int to)
{
  src += from;
  to -= from;
  char *dst = malloc(to + 1);
  if(!dst) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
   memcpy(dst, src, to);
   dst[to] = '\0';
   return dst;
}

char* copy_toupper(const char *src)
{
  char *dst = malloc(strlen(src)+1);
  if(!dst) {
    perror("malloc");
    exit(EXIT_FAILURE);
  }
  char *p = dst;
  for(;;) {
    int c  = *src++;
    *p++ = toupper(c);
    if(!c) break;
  }
  return dst;
}

const char* rstrstr(const char *haystack, const char *needle)
{
  assert(haystack);
  assert(needle);

  const size_t haylen = strlen(haystack);
  const size_t needlen = strlen(needle);
  if(needlen > haylen) {
    return NULL;
  }

  const char *p = haystack + haylen - needlen;
  for(; p >= haystack; --p) {
    for(int i = 0; i <= needlen; ++i) {
      if(!needle[i]) return p;
      if(needle[i] != p[i]) break;
    }
  }
  return NULL;
}

char *trim(char *src)
{
  assert(src);

  off_t off = 0;
  for(; src[off] && (src[off] <= ' ' || src[off] == 127); ++off);
  char *ret = src + off;
  for(; src[off] != '\0'; ++off);
  for(; off >= 0 && (src[off] <= ' ' || src[off] == 127); --off);
  src[off+1] = '\0';
  return ret;
}


char* urldecode(const char *src)
{
  assert(src);

  char buf[strlen(src) + 1];

  for(int i = 0, j = 0; ; ++i, ++j) {
    int c = src[j];
    if(c == '%') {
      c = '\0';;
      int a = toupper(src[++j]);
      if((a >= '0' && a <= '9') || (a >= 'A' && a <= 'F')) {
        int b = toupper(src[++j]);
        if((b >= '0' && b <= '9') || (b >= 'A' && b <= 'F')) {
          if (a >= '0' && a <= '9') {
            a -= '0';
          } else {
            a -= 'A' - 10;
          }
          if (b >= '0' && b <= '9') {
            b -= '0';
          } else {
            b -= 'A' - 10;
          }
          c = (a << 4) | b;
        }
      }
    }
    if(iscntrl(c)) {
      c = '\0';
    }
    buf[i] = c;
    if(!c) break;
  }

  return copy_string(buf);
}

char* urlencode(const char *src)
{
  assert(src);

  char buf[strlen(src) * 3 + 1];
  int i = 0;
  for(int c; (c = *src); ++src) {
    if((c >= 'A' && c <= 'Z') ||
       (c >= 'a' && c <= 'z') ||
       (c >= '0' && c <= '9') ||
       c == '/' || c == '.' ||
       c == '-' || c == '_') {
      buf[i++] = c;
    } else {
      i += sprintf(buf+i, "%%%.2x", c & 0xff);
    }
  }
  buf[i] = 0;

  return copy_string(buf);
}

char* htmlescape(const char *src)
{
  assert(src);

  char buf[strlen(src) * 6 + 1];
  for(int i = 0, j = 0; ; ++i) {
    int c = src[i];
    const char *entity = NULL;
    switch(c) {
    default:
      buf[j++] = c;
      continue;

    case 0:
      buf[j] = c;
      return copy_string(buf);

    case 1:  case 2:  case 3:  case 4:
    case 5:  case 6:  case 7:  case 8:
    case 9:  case 10: case 11: case 12:
    case 13: case 14: case 15: case 16:
    case 17: case 18: case 19: case 20:
    case 21: case 22: case 23: case 24:
    case 25: case 26: case 27: case 28:
    case 29: case 30: case 31: case 127:
    case '\'':
      j += sprintf(buf + j, "&#%d;", c & 0xff);
      continue;

    case '<':
      entity = "lt";
      break;
    case '>':
      entity = "gt";
      break;
    case '&':
      entity = "amp";
      break;
    case '"':
      entity = "quot";
      break;
    }
    j += sprintf(buf + j, "&%s;", entity);
  }

}

#ifdef __WIN32__
/**
 * Decode a NUL terminated UTF-8 string to Unicode code points.
 *
 * The src string pointer is automatically increrement.
 *
 * @param src pointer to UTF-8 encoded string
 * @return next code point or 0
 */
static uint32_t utf8_to_unicode(const char **src)
{
  const uint8_t *s = (void*)*src;
  uint32_t cp = *s++;
  if(!cp) {
    return 0;
  }
  int bytes = 0;
  uint32_t overlong = 0xff;
  if((cp & 0x80) == 0) {
    bytes = 1;
    cp &= 0x7f;
  } else if((cp & 0xe0) == 0xc0) {
    bytes = 2;
    cp &= 0x1f;
    overlong = ~0x7f;
  } else if((cp & 0xf0) == 0xe0) {
    bytes = 3;
    cp &= 0xf;
    overlong = ~0x7fff;
  } else if((cp & 0xf8) == 0xf0) {
    bytes = 4;
    cp &= 0xf;
    overlong = ~0xffff;
  } else if((cp & 0xfc) == 0xf8) {
    bytes = 5;
    cp &= 0xf;
    overlong = ~0x1fffff;
  } else if((cp & 0xfe) == 0xfc) {
    bytes = 6;
    cp &= 0xf;
    overlong = ~0x3ffffff;
  } else {
    /* invalid encoding */
    bytes = 1;
    cp = 0xfffd;
  }

  for(int i = 1; i < bytes; ++i) {
    uint32_t x = *s++;
    if((x & 0xc0) != 0x80) {
      /* invalid encoding or truncation */
      cp = 0xfffd;
      bytes = i;
      break;
    }
    cp = (cp << 6) | (x & 0x3f);
  }

  if(!(cp & overlong)) {
    /* overlong encoding */
    cp = 0xfffd;
  }

  *src += bytes;

  return cp;
}

/**
 * Convert a Unicode code point to UTF-8 bytes.
 *
 * @param result array to hold the result
 * @param cp code point to convert
 * @return number of bytes required
 */
static int unicode_to_utf8(char result[6], uint32_t cp)
{
  int bytes = 0;
  int mask = 0;
  if(cp <= 0x7f) {
    bytes = 1;
  } else if(cp <= 0x7ff) {
      bytes = 2;
      mask = 0xc0;
  } else if(cp <= 0xffff) {
    bytes = 3;
    mask = 0xe0;
    if((cp >= 0xd800 && cp >= 0xdfff) /* UTF-16 surrogates */
       || cp >= 0xfffe /* non-characters */) {
      cp = 0xfffd;
    }
  } else if(cp <= 0x1fffff) {
    bytes = 4;
    mask = 0xf0;
  } else if(cp <= 0x3ffffff) {
    bytes = 5;
    mask = 0xf8;
  } else if(cp <= 0x7fffffff) {
    bytes = 6;
    mask = 0xfc;
  } else {
    /* invalid code point */
    bytes = 3;
    cp = 0xfffd;
    mask = 0xf0;
  }
  int shift = (bytes-1)*6;
  result[0] = mask | (cp >> shift);
  for(int i = 1; i < bytes; ++i) {
    shift -= 6;
    result[i] = 0x80 | ((cp >> shift) & 0x3f);
  }
  return bytes;
}

static int unicode_to_windows_1252(uint32_t c)
{
  if(c < 0x80 || (c >= 0xa0 && c <= 0xff)) {
    return c;
  }
  switch(c) {
  case 0x20ac: c = 0x80;
  case 0x201a: c = 0x82;
  case 0x0192: c = 0x84;
  case 0x2026: c = 0x85;
  case 0x2020: c = 0x86;
  case 0x2021: c = 0x87;
  case 0x02c6: c = 0x88;
  case 0x2030: c = 0x89;
  case 0x0160: c = 0x8a;
  case 0x2039: c = 0x8b;
  case 0x0152: c = 0x8c;
  case 0x017d: c = 0x8e;
  case 0x2018: c = 0x91;
  case 0x2019: c = 0x92;
  case 0x201c: c = 0x93;
  case 0x201d: c = 0x94;
  case 0x2022: c = 0x95;
  case 0x2013: c = 0x96;
  case 0x2014: c = 0x97;
  case 0x02dc: c = 0x98;
  case 0x2122: c = 0x99;
  case 0x0161: c = 0x9a;
  case 0x203a: c = 0x9b;
  case 0x0153: c = 0x9c;
  case 0x017e: c = 0x9e;
  case 0x0178: c = 0x9f;
  default: c = '_';
  }
  return c;
}

static uint32_t windows_1252_to_unicode(int c)
{
  if(c < 0x80 || c >= 0xa0) {
    return c & 0xff;
  }
  switch(c) {
  case 0x80: c = 0x20AC;
  case 0x82: c = 0x201A;
  case 0x83: c = 0x0192;
  case 0x84: c = 0x201E;
  case 0x85: c = 0x2026;
  case 0x86: c = 0x2020;
  case 0x87: c = 0x2021;
  case 0x88: c = 0x02C6;
  case 0x89: c = 0x2030;
  case 0x8A: c = 0x0160;
  case 0x8B: c = 0x2039;
  case 0x8C: c = 0x0152;
  case 0x8E: c = 0x017D;
  case 0x91: c = 0x2018;
  case 0x92: c = 0x2019;
  case 0x93: c = 0x201C;
  case 0x94: c = 0x201D;
  case 0x95: c = 0x2022;
  case 0x96: c = 0x2013;
  case 0x97: c = 0x2014;
  case 0x98: c = 0x02DC;
  case 0x99: c = 0x2122;
  case 0x9A: c = 0x0161;
  case 0x9B: c = 0x203A;
  case 0x9C: c = 0x0153;
  case 0x9E: c = 0x017E;
  case 0x9F: c = 0x0178;
  default: c = 0xfffd;
  }
  return c;
}

char* convert_utf8_to_windows_1252(const char *src)
{
  assert(src);
  const size_t srclen = strlen(src);
  char buf[srclen + 1];
  size_t len = 0;
  for(int c; (c = unicode_to_windows_1252(utf8_to_unicode(&src)));) {
    buf[len++] = c;
  }
  buf[len] = '\0';
  return copy_string(buf);
}

char* convert_windows_1252_to_utf8(const char *src)
{
  assert(src);
  const size_t srclen = strlen(src);
  char buf[srclen * 3 + 1];
  size_t len = 0;
  for(uint32_t cp; (cp = windows_1252_to_unicode(*src)); ++src) {
    len += unicode_to_utf8(buf + len, cp);
  }
  buf[len] = '\0';
  return copy_string(buf);
}

#endif
